/*
 * Licensed to STRATIO (C) under one or more contributor license agreements.
 * See the NOTICE file distributed with this work for additional information
 * regarding copyright ownership.  The STRATIO (C) licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package com.stratio.crossdata.common.logicalplan;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.stratio.crossdata.common.metadata.ColumnType;
import com.stratio.crossdata.common.metadata.Operations;
import com.stratio.crossdata.common.statements.structures.FunctionSelector;
import com.stratio.crossdata.common.statements.structures.Selector;

/**
 * Select operator that specifies the list of columns that should be returned with their expected name. Notice that this
 * operator is applied to limit the number of columns returned and to provide alias support.
 */
public class Select extends TransformationStep {

    private static final long serialVersionUID = -2816531584230162082L;
    /**
     * Map of columns associating the name given in the Project logical steps with the name expected in the results.
     */
    private Map<Selector, String> columnMap;

    /**
     * Map of {@link com.stratio.crossdata.common.metadata.ColumnType} associated with each column.
     */
    private Map<String, ColumnType> typeMap;

    /**
     * Map with the types of the selectors.
     */
    private Map<Selector, ColumnType> typeMapFromColumnName;

    private boolean isDistinct = false;

    /**
     * Class constructor.
     *
     * @param requiredOperations
     *            A set of operations to be applied.
     * @param columnMap
     *            Map of columns associating the name given in the Project logical steps with the name expected in the
     *            result.
     * @param typeMap
     *            The mapping of column types.
     * @param typeMapFromColumnName  The types of selectors.
     * @param isDistinct Whether the select expression is modified by the operator distinct or not.
     */
    public Select(Set<Operations> requiredOperations, Map<Selector, String> columnMap, Map<String, ColumnType> typeMap,
            Map<Selector, ColumnType> typeMapFromColumnName, boolean isDistinct) {
        super(requiredOperations);
        this.columnMap = columnMap;
        this.typeMap = typeMap;
        this.typeMapFromColumnName = typeMapFromColumnName;
        this.isDistinct = isDistinct;
    }

    /**
     * Class constructor.
     * 
     * @param requiredOperations
     *            A set of operations to be applied.
     * @param columnMap
     *            Map of columns associating the name given in the Project logical steps with the name expected in the
     *            result.
     * @param typeMap
     *            The mapping of column types.
     * @param typeMapFromColumnName  The types of selectors.
     */
    public Select(Set<Operations> requiredOperations, Map<Selector, String> columnMap, Map<String, ColumnType> typeMap,
            Map<Selector, ColumnType> typeMapFromColumnName) {
        this(requiredOperations, columnMap, typeMap, typeMapFromColumnName, false);
    }

    public Map<Selector, String> getColumnMap() {
        return columnMap;
    }

    public Map<String, ColumnType> getTypeMap() {
        return typeMap;
    }

    public Map<Selector, ColumnType> getTypeMapFromColumnName() {
        return typeMapFromColumnName;
    }

    public void setColumnMap(Map<Selector, String> columnMap) {
        this.columnMap = columnMap;
    }

    public void setTypeMap(Map<String, ColumnType> typeMap) {
        this.typeMap = typeMap;
    }

    public void setTypeMapFromColumnName(Map<Selector, ColumnType> typeMapFromColumnName) {
        this.typeMapFromColumnName = typeMapFromColumnName;
    }

    public boolean isDistinct() {
        return isDistinct;
    }

    public void setIsDistinct(boolean isDistinct) {
        this.isDistinct = isDistinct;
    }

    /**
     * Get the selectors that must be returned as a result of executing the query
     * in the expected order.
     * @return A list of {@link com.stratio.crossdata.common.statements.structures.Selector}.
     */
    public List<Selector> getOutputSelectorOrder(){
        List<Selector> result = new ArrayList<>();
        for(Selector selector: columnMap.keySet()){
            result.add(selector);
        }
        return result;
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder("SELECT (");
        if(isDistinct){
            sb.append("DISTINCT ");
        }
        Iterator<Map.Entry<Selector, String>> it = columnMap.entrySet().iterator();
        Map.Entry<Selector, String> entry;
        while (it.hasNext()) {
            entry = it.next();
            Selector key = entry.getKey();
            if(FunctionSelector.class.isInstance(key)){
                FunctionSelector fs = FunctionSelector.class.cast(key);
                if(fs.getFunctionColumns().isDistinct()){
                    sb.append("DISTINCT ");
                }
            }
            sb.append(key.getColumnName().getQualifiedName()).append(" AS ").append(entry.getValue());
            if (it.hasNext()) {
                sb.append(", ");
            }
        }
        sb.append(")");
        return sb.toString();
    }
}
