/*
 * Licensed to STRATIO (C) under one or more contributor license agreements.
 * See the NOTICE file distributed with this work for additional information
 * regarding copyright ownership.  The STRATIO (C) licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package com.stratio.crossdata.core.grid;

import static org.testng.Assert.assertTrue;

import java.util.Map;
import java.util.concurrent.locks.Lock;

import javax.transaction.TransactionManager;

import org.jgroups.JChannel;
import org.jgroups.Message;
import org.jgroups.ReceiverAdapter;
import org.jgroups.blocks.MessageDispatcher;
import org.jgroups.blocks.RequestHandler;
import org.jgroups.blocks.RequestOptions;
import org.testng.Assert;
import org.testng.annotations.AfterClass;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

import com.stratio.crossdata.common.exceptions.ManifestException;
import com.stratio.crossdata.core.MetadataManagerTestHelper;

/**
 * Tests {@link Grid}.
 */
@Test(testName = "GridTest")
public class GridTest {

    private String syncMessage;
    private String asyncMessage;
    private String path;

    @BeforeClass
    public void setUp() throws ManifestException {
        MetadataManagerTestHelper.HELPER.initHelper();
        MetadataManagerTestHelper.HELPER.createTestEnvironment();
    }

    @AfterClass
    public void tearDown() throws Exception {
        MetadataManagerTestHelper.HELPER.closeHelper();
    }

    /**
     * Tests {@link Grid} distributed storing.
     */
    @Test
    public void testGridStore() throws Exception {
        Map<String, String> map = Grid.INSTANCE.map("testGridStore");
        TransactionManager tm = Grid.INSTANCE.transactionManager("testGridStore");
        tm.begin();
        Assert.assertNotNull(map);
        map.put("k1", "v1");
        Assert.assertEquals(map.get("k1"), "v1");
        map.remove("k1");
        Assert.assertNull(map.get("k1"));
        tm.commit();
    }

    /**
     * Tests {@link Grid} distributed locking.
     */
    @Test
    public void testGridLock() throws Exception {
        boolean res = true;
        Lock lock = Grid.INSTANCE.lock("testGridLock");
        try {
            lock.lock();
        } finally {
            lock.unlock();
        }
        assertTrue(res, "Grid Lock test failed.");
    }

    /**
     * Tests {@link Grid} distributed synchronous channeling.
     */
    @Test
    public void testGridSyncChannel() throws Exception {
        JChannel syncChannel = Grid.INSTANCE.channel("testGridSyncChannel");
        MessageDispatcher
                dispatcher =
                new MessageDispatcher(syncChannel, null, null, new RequestHandler() {
                    @Override
                    public Object handle(Message msg) throws Exception {
                        syncMessage = msg.getObject().toString();
                        return msg;
                    }
                });
        syncChannel.connect("test");
        dispatcher.castMessage(null, new Message(null, "hello"), RequestOptions.SYNC());
        Assert.assertEquals(syncMessage, "hello");
    }

    /**
     * Tests {@link Grid} distributed asynchronous channeling.
     */
    @Test
    public void testGridAsyncChannel() throws Exception {
        JChannel asyncChannel = Grid.INSTANCE.channel("testGridAsyncChannel");
        asyncChannel.setReceiver(new ReceiverAdapter() {
            public void receive(Message msg) {
                asyncMessage = msg.getObject().toString();
            }
        });
        asyncChannel.connect("test");
        asyncChannel.send(new Message(null, "hello"));
        Thread.sleep(2000);
        Assert.assertEquals(asyncMessage, "hello");
    }

}
